@@ -159,8 +159,36 @@ class _TestEtcdBase(lib.Lib, unittest.TestCase):
             self.assertEqual(expected, self.recent_deletes)
         self.recent_deletes = set()
 
-    def etcd3_get(self, key, metadata=False):
+    def etcd3_get(
+        self,
+        key,
+        metadata=False,
+        range_end=None,
+        sort_target=None,
+        sort_order=None,
+        limit=None,
+        revision=None,
+    ):
         self.maybe_reset_etcd()
+
+        if range_end is not None:
+            # Ranged get...
+            decoded_end = _decode(range_end)
+            _log.info("Ranged get %s...%s", key, decoded_end)
+            assert revision is not None
+            keys = self.etcd_data.keys()
+            keys.sort()
+            if sort_order == "descend":
+                keys.reverse()
+            result = []
+            keys_in_range = [k for k in keys if key <= k < decoded_end]
+            for k in keys_in_range:
+                result.append((self.etcd_data[k],
+                               {'key': k, 'mod_revision': '10'}))
+                if limit is not None and len(result) >= limit:
+                    break
+            return result
+
         if key in self.etcd_data:
             value = self.etcd_data[key]
 
@@ -283,11 +311,12 @@ class TestPluginEtcd(_TestEtcdBase):
 
         # Mock out config.
         lib.m_compat.cfg.CONF.calico.etcd_host = "localhost"
-        lib.m_compat.cfg.CONF.calico.etcd_port = 4001
+        lib.m_compat.cfg.CONF.calico.etcd_port = 2379
         lib.m_compat.cfg.CONF.calico.etcd_cert_file = None
         lib.m_compat.cfg.CONF.calico.etcd_ca_cert_file = None
         lib.m_compat.cfg.CONF.calico.etcd_key_file = None
         lib.m_compat.cfg.CONF.calico.num_port_status_threads = 4
+        lib.m_compat.cfg.CONF.calico.etcd_compaction_period_mins = 0
 
     sg_default_key_v3 = (
         '/calico/resources/v3/projectcalico.org/networkpolicies/' +
